package icse2011experiments;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import ca.mcgill.cs.swevo.diffcat.model.DiffCatResult;
import ca.mcgill.cs.swevo.diffcat.model.DiffCodes;
import ca.mcgill.cs.swevo.diffcat.util.filter.IsChangeOperationFilter;
import ca.mcgill.cs.swevo.semdiff.core.Detector;
import ca.mcgill.cs.swevo.semdiff.core.DetectorInput;
import ca.mcgill.cs.swevo.semdiff.core.DetectorResults;
import ca.mcgill.cs.swevo.semdiff.core.Repository;
import ca.mcgill.cs.swevo.semdiff.core.Result;
import ca.mcgill.cs.swevo.semdiff.core.Transaction;
import ca.mcgill.cs.swevo.util.CollectionUtil;
import ca.mcgill.cs.swevo.util.LCRTuple;

/**
 * Used to conduct experiments for ICSE paper "Non-Essential Differences in Version Histories"
 * 
 * @author dkawry
 */

public class EntryPoint implements Detector {

	/**
	 * Modify this to record results for different projects
	 */
	public static final String 	CUR_PROJ = "hibernate";
	
	/**
	 * File used to record Churn statistics (Table 2 in paper)
	 */
	public static final String	LOC_FILE = "/home/2005/dkawry/research/experiments/icse11/loc/" + CUR_PROJ + "/results.txt";
	
	/**
	 * File used to record method update statistics (Table 3 in paper)
	 */
	public static final String	METHOD_FILE = "/home/2005/dkawry/research/experiments/icse11/method/" + CUR_PROJ + "/results.txt";
	
	/**
	 * File used to record how often each kind of non-essential difference appears
	 */
	public static final String  TYPE_FILE = "/home/2005/dkawry/research/experiments/icse11/type/" + CUR_PROJ + "/results.txt";
	
	/**
	 * File used to record the methods modified by each change set
	 */
	public static final String	WEKA_ALL_FILE = "/home/2005/dkawry/research/experiments/icse11/weka/" + CUR_PROJ + "/all.txt";
	
	/**
	 * File used to record the essential methods modified by each change set
	 */
	public static final String	WEKA_ESS_FILE = "/home/2005/dkawry/research/experiments/icse11/weka/" + CUR_PROJ + "/ess.txt";
	
	private final String	MAIN_SEP = "#";
	
	/**
	 * Writer to write churn stats
	 */
	private PrintWriter locWriter;
	
	/**
	 * Writer to write method update stats
	 */
	private PrintWriter methodWriter;
	
	/**
	 * To collect NPEs
	 */
	private PrintWriter bugWriter;
	
	/**
	 * Writer to write kinds of diffs
	 */
	private PrintWriter typeWriter;
	
	private List<String> nonEssentials;
	
	/**
	 * To write all method updates per change set
	 */
	private final ItemSetRecorder allItemSets;
	
	/**
	 * To write all essential updates per change set
	 */
	private final ItemSetRecorder essentialItemSets;
	
	/**
	 * To record the kinds of diffs we encounter
	 */
	private final DiffTypeCounter diffTypeCounter;
	
	public EntryPoint () {
		
		allItemSets = new ItemSetRecorder(WEKA_ALL_FILE);
		essentialItemSets = new ItemSetRecorder(WEKA_ESS_FILE);
		
		diffTypeCounter = new DiffTypeCounter();
		
		 try {
			nonEssentials = new ArrayList<String>(DiffCodes.NON_ESSENTIAL_DIFF.getOps());
			Collections.sort(nonEssentials);
			
			locWriter = new PrintWriter(new File(LOC_FILE));
			methodWriter = new PrintWriter(new File(METHOD_FILE));
			typeWriter = new PrintWriter(new File(TYPE_FILE));
			locWriter.print("num" + MAIN_SEP + "all");
			methodWriter.print("num" + MAIN_SEP + "all");
			
			for (String nonEssential : nonEssentials) {
				locWriter.print(MAIN_SEP + nonEssential + "- ~ +");
				methodWriter.print(MAIN_SEP + nonEssential);
			}
			
			methodWriter.print(MAIN_SEP + "none");
			
			locWriter.println();
			methodWriter.println();
			
			locWriter.flush();
			methodWriter.flush();
			
			bugWriter = new PrintWriter(new File("/home/2005/dkawry/research/experiments/icse11/bugs/" + CUR_PROJ + "/recordBugs.txt"));
			
		} catch (FileNotFoundException e) {
			
			e.printStackTrace();
		}
	}
	
	@Override
	public DetectorResults detect(Repository repository, Transaction transaction, DetectorInput input) {

		try {
			List<LCRTuple<Integer>> locStats 	= new ArrayList<LCRTuple<Integer>>();
			List<Integer> updateStats 			= new ArrayList<Integer>();
			
			// get diffcat results
			List<DiffCatResult> diffCatResults = convert(input.getRequiredResults("ca.mcgill.cs.swevo.diffcat"));
			
			// print trxn number 
			locWriter.print(transaction.getId() + MAIN_SEP);
			methodWriter.print(transaction.getId() + MAIN_SEP);
			
			// get all diff types
			diffTypeCounter.record(transaction.getId(), diffCatResults, typeWriter);
			
			// do loc computation w/o removing any non-essential diffs
			locStats.add(LocCalculator.computeFileLocMetrics(diffCatResults));
			
			// do method update computation w/o removing any non-essential diffs
			Collection<DiffCatResult> copy = CollectionUtil.filterAndCopy(diffCatResults, DiffCodes.METHOD_BODY_DIFF);
			MethodUpdateCounter methodUpdateCounter = new MethodUpdateCounter(copy);
			updateStats.add(methodUpdateCounter.getMethodKeys().getSize());
			
			// remove each non-essential diff one by one and re-do stats
			for (String nonEssentialDiff : nonEssentials) {
				IsChangeOperationFilter filter = new IsChangeOperationFilter(nonEssentialDiff);
				
				Collection<DiffCatResult> toRemove = 
					CollectionUtil.filterAndCopy(diffCatResults, filter);
				
				Collection<DiffCatResult> leftOver = new ArrayList<DiffCatResult>(diffCatResults);
				leftOver.removeAll(toRemove);
				
				locStats.add(LocCalculator.computeFileLocMetrics(leftOver));
				
				// Method Updates
				updateStats.add(methodUpdateCounter.getNumNonEssentialWithChangeType(DiffCodes.NON_ESSENTIAL_DIFF, filter));
			}
			
			// do method updates for all non-essentials
			updateStats.add(methodUpdateCounter.getNumNonEssentialKeys(DiffCodes.NON_ESSENTIAL_DIFF));
			
			// print loc stats
			printLocStats(locStats);
			
			// print method stats
			printMethodStats(updateStats);
			
			// add to weka experiment
			allItemSets.recordAll(transaction.getId(), diffCatResults);
			essentialItemSets.recordAll(transaction.getId(), CollectionUtil.filterAndCopy(diffCatResults, DiffCodes.ESSENTIAL_DIFF));
		} catch (Exception e) {
			bugWriter.println("Failed on " + transaction.getId());
			e.printStackTrace(bugWriter);
			bugWriter.flush();
		}
		
		return new DetectorResults();
	}
	
	private void printMethodStats(List<Integer> updateStats) {
		// print the method update results to file
		for (Integer num : updateStats) {
			methodWriter.print(num + MAIN_SEP);
		}
		methodWriter.println();
		methodWriter.flush();
	}

	private void printLocStats(List<LCRTuple<Integer>> locStats) {
		// print the loc results to file
		for (LCRTuple<Integer> locStat: locStats) {
			locWriter.print(locStat.toString() + MAIN_SEP);
		}
		locWriter.println();
		locWriter.flush();
	}

	private List<DiffCatResult> convert(List<Result> results) {
		List<DiffCatResult> diffCatResults = new LinkedList<DiffCatResult>();
		for (Result result : results){
			diffCatResults.add((DiffCatResult) result);
		}
		return diffCatResults;
	}
}
