package icse2011experiments;

import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.AbstractTypeDeclaration;
import org.eclipse.jdt.core.dom.FieldDeclaration;
import org.eclipse.jdt.core.dom.MethodDeclaration;

import ca.mcgill.cs.swevo.diffcat.model.ChangeOperation;
import ca.mcgill.cs.swevo.diffcat.model.DiffCatResult;
import ca.mcgill.cs.swevo.diffcat.model.DiffCodes;
import ca.mcgill.cs.swevo.diffcat.util.ast.ASTUtil;
import ca.mcgill.cs.swevo.diffcat.util.filter.IsChangeOperationFilter;
import ca.mcgill.cs.swevo.diffcat.util.filter.NullElimFilter;
import ca.mcgill.cs.swevo.util.CollectionUtil;
import ca.mcgill.cs.swevo.util.LCRTuple;
import ca.mcgill.cs.swevo.util.LRTuple;
import ca.mcgill.cs.swevo.util.filter.IFilter;

/**
 * This file computes the loc statistics for Table 2 of David 
 * and Martin's ICSE 2011 submission.  
 *  
 * The {@link LocCalculator} is used by the {@link EntryPoint}
 * class, the main class of the icse2011Experiments project.  
 * 
 * Comments within the code explain how we do it. 
 * 
 * @author dkawry
 */
public class LocCalculator {

	public static LCRTuple<Integer> computeFileLocMetrics (Collection<DiffCatResult> diffs)
	{
		int d=0, u=0, i=0;
		
		diffs = CollectionUtil.filterAndCopy(diffs, new NullElimFilter<DiffCatResult>());
		
		// count class signature update 
		// (1loc per class signature update, even if spans multiple lines)
		u += getLocUpdates(diffs, DiffCodes.CLASS_SIGNATURE_DIFF);
		
		// count field signature updates 
		// (1loc per field signature update, even if spans multiple lines)
		u += getLocUpdates(diffs, DiffCodes.FIELD_LEVEL_DIFF);
		
		// count method signature updates
		// (1loc per field signature update, even if spans multiple lines)
		u += getLocUpdates(diffs, DiffCodes.METHOD_SIGNATURE_DIFF);
		
		// get locs for attribute, method, and class inserts and deletes
		LRTuple<Integer> info = getRawChunkInserts(diffs);
		d += info.getLeft();
		i += info.getRight();
		
		// get locs for intra-method code chunks
		LCRTuple<Integer> hunks = getHunkMetrics(diffs);
		d += hunks.getLeft();
		u += hunks.getCenter();
		i += hunks.getRight();
		
		return new LCRTuple<Integer> (d, u, i);
	}
	
	/**
	 * Helper method. Computes 1loc per diff passed into this method. 
	 * The filter collects those diffs that should be counted. 
	 */
	private static int getLocUpdates (Collection<DiffCatResult> diffs, IFilter<DiffCatResult> filter)
	{
		Set<ASTNode> helperMap = new HashSet<ASTNode>();
		for (DiffCatResult diff : diffs) {
			if (filter.passesFilter(diff)) {
				helperMap.add(diff.getLeftParent());
			}
		}
		
		return helperMap.size();
	}
	
	/**
	 * Helper method counting the deleted and added lines for a given collection of 
	 * differences 
	 */
	private static LRTuple<Integer> getRawChunkInserts (Collection<DiffCatResult> diffs)
	{
		int locDeleted = 0;
		int locAdded = 0;
		
		for (DiffCatResult diff : diffs) {
			if (DiffCodes.ELEMENT_LEVEL_DIFF.passesFilter(diff) ||
				DiffCodes.CLASS_INS_DEL_DIFF.passesFilter(diff)) {
				if (diff.isDelete()) {
					locDeleted += getNumLines (diff.getNodeFrom());
				}
				if (diff.isInsert()) {
					locAdded += getNumLines(diff.getNodeTo());
				}
			}
		}
		
		return new LRTuple<Integer>(locDeleted, locAdded);
	}
	
	/**
	 * Helper method to compute locs for intra-method differences. 
	 */
	private static LCRTuple<Integer> getHunkMetrics (Collection<DiffCatResult> diffs) 
	{
		int locDeleted = 0;
		int locUpdated = 0;
		int locAdded = 0;
		
		diffs = CollectionUtil.filterAndCopy(diffs, DiffCodes.METHOD_BODY_DIFF);
		
		// take care of variable declaration statements (b/c multi-var declarations
		// are counted separately, but are on the same line)
		Set<ASTNode> varDeclStatements = new HashSet<ASTNode>();
		Collection<DiffCatResult> varDeclDiffs = CollectionUtil.filterAndCopy(diffs, new VarDeclStmtFilter());
		
		for (DiffCatResult diff : varDeclDiffs) 
		{
			if (diff.getNodeTo() == null) {
				continue;
			}
			ASTNode parent = ASTUtil.getParentWithType(ASTNode.VARIABLE_DECLARATION_STATEMENT, diff.getNodeTo());
			if (parent == null) {
				continue;
			}
			varDeclStatements.add(parent);
		}
		
		// we just group them based on their parent stmts
		for (ASTNode varDecl : varDeclStatements) {
			locUpdated += getNumLines(varDecl.toString());
		}
		
		// otherwise do generic handling
		diffs.removeAll(varDeclDiffs);
		for (DiffCatResult diff : diffs) {
			
			if (diff.isDelete()) {
				if (diff.getNodeFrom() == null) {
					locDeleted += 1;	// for unclassified diffs
				}
				else {
					locDeleted += getNumLines(diff.getNodeFrom().toString());
				}
			}
			if (diff.isInsert()) {
				if (diff.getNodeTo() == null) {
					locAdded += 1;		// for unclassified diffs
				} else {
					locAdded += getNumLines (diff.getNodeTo().toString());
				}
			}
			// treat moves as del/add pair 
			if (diff.isMove()) {
				locDeleted += diff.getNodeFrom() != null ? getNumLines (diff.getNodeFrom().toString()) : 1;
				locAdded += diff.getNodeTo() != null ? getNumLines (diff.getNodeTo().toString()) : 1;
			}
			if (diff.isUpate()) {
				locUpdated += diff.getNodeTo() != null ? getNumLines (diff.getNodeTo().toString()) : 1;
			}
		}
		
		return new LCRTuple<Integer>(locDeleted, locUpdated, locAdded);
	}
	
	/**
	 * Helper method computing the loc for a given AST node
	 */
	private static int getNumLines (ASTNode node) {
		
		if (node instanceof MethodDeclaration) {
			return getNumLines ((MethodDeclaration) node);
		}
		if (node instanceof FieldDeclaration) {
			return getNumLines ((FieldDeclaration) node);
		}
		if (node instanceof AbstractTypeDeclaration) {
			return getNumLines ((AbstractTypeDeclaration) node);
		}
		
		return 0;
	}
	
	/**
	 * Helper method adding the loc for a method declaration (plus body)
	 * we do +1 for the signature, and then we do the body. 
	 */
	private static int getNumLines (MethodDeclaration decl) {
		if (decl.getBody() == null) {
			return 1;
		}
		int counter = 1 + getNumLines (decl.getBody().toString());	// one for method signature
		return counter;
	}
		
	private static int getNumLines (FieldDeclaration decl) {
		return getNumLines(decl.toString());
	}
	
	/**
	 * Helper method computing the loc for a type declaration (plus body)
	 * Delegates for all methods, fields, and member types (i.e., ignores 
	 * static blocks and white space) 
	 */
	private static int getNumLines (AbstractTypeDeclaration decl) {
		
		Set<ASTNode> memberTypes = ASTUtil.getMemberTypes(decl);
		List<MethodDeclaration> methods = ASTUtil.getMethods(decl);
		List<FieldDeclaration> fields = ASTUtil.getFields(decl);
		
		int numLines = getNumLinesSignature(decl);
		
		for (FieldDeclaration fieldDecl : fields) {
			numLines += getNumLines (fieldDecl);
		}
		
		for (MethodDeclaration methodDecl : methods) {
			numLines += getNumLines (methodDecl);
		}
		
		for (ASTNode node : memberTypes) {
			numLines += getNumLines((AbstractTypeDeclaration) node);
		}
		
		return numLines;
	}
	
	private static int getNumLinesSignature (AbstractTypeDeclaration decl) {
		return 1;
	}
	
	/**
	 * Counts the lines of code for a given (code) string 
	 * We ignore { and } when computing code lines. 
	 */
	private static int getNumLines (String s) {
		String[] a = s.split("\n");
		int c = 0;
		for (String ai : a) {
			ai.trim();
			ai = ai.replaceAll("\\{", "");
			ai = ai.replaceAll("\\}", "");
			if (ai.isEmpty()) {
				continue;
			}
			c ++;
		}
		return c;
	}
	
	private static class VarDeclStmtFilter extends IsChangeOperationFilter {
		public VarDeclStmtFilter () {
			super(
				ChangeOperation.VARIABLE_DECLARED_TYPE_UPDATE.name(),
				ChangeOperation.VARIABLE_DECLARED_TYPE_TRIVIAL_UPDATE.name(),
				ChangeOperation.VARIABLE_INITIALIZE_DELETE.name(),
				ChangeOperation.VARIABLE_INITIALIZE_INSERT.name(),
				ChangeOperation.VARIABLE_INITIALIZE_UPDATE.name(),
				ChangeOperation.VARIABLE_NAME_UPDATE.name());
		}
	}
}
