package icse2011experiments;

import java.util.Collection;
import java.util.LinkedList;

import org.eclipse.jdt.core.dom.ASTNode;

import ca.mcgill.cs.swevo.diffcat.model.DiffCatResult;
import ca.mcgill.cs.swevo.diffcat.util.DiffCatParentMap;
import ca.mcgill.cs.swevo.util.CollectionUtil;
import ca.mcgill.cs.swevo.util.LRTuple;
import ca.mcgill.cs.swevo.util.LRTupleSet;
import ca.mcgill.cs.swevo.util.filter.IFilter;
 
/**
 * Helper class. Used to produce information for files within 
 * the method/project directories of David and Martin's ICSE 
 * 2011 submission. 
 * 
 * Comments within the file explain how we did it. 
 */
public class MethodUpdateCounter {

	/**
	 * Used to store and sort all DiffCatResults according to their 
	 * parents (methods or types)
	 */
	private final DiffCatParentMap map;
	
	public MethodUpdateCounter (Collection<DiffCatResult> results) {
		map = new DiffCatParentMap(results);
	}
	
	/**
	 * @return all method pairs containing differences
	 */
	public LRTupleSet<ASTNode> getMethodKeys () {
		return map.getMethodKeys();
	}
	
	/**
	 * Computes the number of method pairs featuring only the kinds of differences
	 * matching isNonEssential
	 * @param isNonEssential A filter to determine the kinds of differences to consider
	 * 		  non-essential. 
	 */
	public int getNumNonEssentialKeys (IFilter<DiffCatResult> isNonEssential) {
		LRTupleSet<ASTNode> all = getMethodKeys();
		int num = 0;
		for (LRTuple<ASTNode> pair : all) {
			if (isNonEssential(pair, isNonEssential)) {
				num++;
			}
		}
		return num;
	}
	
	/**
	 * Computes all non-essential method pairs that feature at least one difference
	 * of the given kind. 
	 * @param isNonEssential Filter determining whether method features only non-essentials
	 * @param toMatch Filter specifying the kind of change type being checked. 
	 * @return
	 */
	public int getNumNonEssentialWithChangeType (IFilter<DiffCatResult> isNonEssential, IFilter<DiffCatResult> toMatch) {
		LRTupleSet<ASTNode> all = getMethodKeys();
		int num = 0;
		for (LRTuple<ASTNode> pair : all) {
			// is method non-essential?
			if (! isNonEssential(pair, isNonEssential)) {
				continue;
			}
			
			Collection<DiffCatResult> diffs = map.get(pair);
			
			// if so, check if at least one diff is of the given type
			if (CollectionUtil.contains(diffs, toMatch)) {
				num ++;
			}
		}
		
		return num;
	}
	
	/**
	 * Checks whether all the differences for the given method pair methodPair are non-essential.
	 * @param methodPair The method pair being checked
	 * @param isNonEssential Filter determining whether a given difference is non-essential
	 * @return Whether methodPair has only non-essential differences. 
	 */
	public boolean isNonEssential (LRTuple<ASTNode> methodPair, IFilter<DiffCatResult> isNonEssential) {
		
		Collection<DiffCatResult> updates = map.get(methodPair);
		Collection<DiffCatResult> copy = new LinkedList<DiffCatResult>(updates);
		copy = CollectionUtil.removeAll(copy, isNonEssential);
		
		return copy.isEmpty();
	}
}
