package icse2011experiments.assoc;

import java.io.File;
import java.io.PrintWriter;
import java.security.AllPermission;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import ca.mcgill.cs.swevo.diffcat.util.binding.BindingUtil;
import ca.mcgill.cs.swevo.util.CollectionMap;
import ca.mcgill.cs.swevo.util.CollectionUtil;
import ca.mcgill.cs.swevo.util.LRTuple;
import ca.mcgill.cs.swevo.util.LRTupleSet;
import ca.mcgill.cs.swevo.util.ListUtil;
import ca.mcgill.cs.swevo.util.MapUtil;
import ca.mcgill.cs.swevo.util.SetUtil;
import ca.mcgill.cs.swevo.util.filter.IFilter;

/**
 * 
 * This file was used to perform the mini association experiment in section 5.3 of the ICSE
 * paper. The file takes as input (for each project) the allR.txt and essR.txt files containing
 * *all* or only the *essential* method updates per transaction, respectively. As output, it 
 * computes, for each *essential* method update per transaction, the sorted recommendations 
 * that would have been made by a Zimmermann-et-al-style association ruling analysis based on 
 * the previous method updates (either all or essential). Comments within the file explain 
 * how this is done.
 */
public class AssociationChecker {

	public static void main(String[] args) {
		
		/**
		 * Max number of suggestions made per method update
		 */
		final int SUGG_CUTOFF = 10;
		
		try {
			final String[] projects = new String[] { "spring", "vuze", "hibernate", "jdt_core", "jdt_ui",  "ant", "xerces"}; 
			
			PrintWriter pwAll  = new PrintWriter(new File("/home/2005/dkawry/research/experiments/icse11/weka/allAnalysis.txt"));
			PrintWriter pwEss  = new PrintWriter(new File("/home/2005/dkawry/research/experiments/icse11/weka/essAnalysis.txt"));

			for (String project : projects) {
				
				// obtain raw 'all' method update info
				final String readFromAll = "/home/2005/dkawry/research/experiments/icse11/weka/" + project + "/allR.txt";
				Map<Integer, Set<String>> transactionsAll = getInfo(readFromAll);
				
				// obtain raw 'ess' method update info
				final String readFromEss = "/home/2005/dkawry/research/experiments/icse11/weka/" + project + "/essR.txt";
				Map<Integer, Set<String>> transactionsEss = getInfo(readFromEss);
				
				// all trxns
				List<Integer> trxnsAll = MapUtil.getKeysSorted(transactionsAll);
				
				/**
				 * Records all the transactions that a given method pair appeared in
				 */
				final CollectionMap<LRTuple<String>, Integer> pairsAll = new CollectionMap<LRTuple<String>, Integer>();
				
				/**
				 * Records all the transactions that a given method pair appeared in modulo non-essential updates
				 */
				final CollectionMap<LRTuple<String>, Integer> pairsEss = new CollectionMap<LRTuple<String>, Integer>();
				
				/**
				 * Records all the transactions that a given method appears in 
				 */
				final CollectionMap<String, Integer> rawCounterAll = new CollectionMap<String, Integer>();
				
				/**
				 * Records all the transactions that a given method pair appeared in modulo non-essential updates
				 */
				final CollectionMap<String, Integer> rawCounterEss = new CollectionMap<String, Integer>();
				
				// do experiment w.r.t. all method updates
				for (int trxn : trxnsAll) {
					
					if ( (trxn % 100) == 0) {
						System.out.println("Doing: " + trxn);
					}
					
					Set<String> methodsAll = transactionsAll.get(trxn);
					Set<String> methodsEss = transactionsEss.get(trxn);
					
					if (methodsEss == null) {
						methodsEss = new HashSet<String>();
					}
					
					CollectionMap<String,LRTuple<String>> pairs = collectPosPairs(pairsAll, methodsEss);
					
					// each method is a "navigation scenario" a la ZimmermannEvaluator et al
					for (String methodEss : methodsEss) 
					{
						// get all possible pairs for this method seed
						List<LRTuple<String>> posPairs = new ArrayList<LRTuple<String>>(pairs.get(methodEss));
						
						final Map<LRTuple<String>, Double> confidences = new HashMap<LRTuple<String>, Double>();
						final Map<LRTuple<String>, Double> supports = new HashMap<LRTuple<String>, Double>();
						
						// compute supports and confidences 
						for (LRTuple<String> posPair : posPairs) {
							confidences.put(posPair, computeConfidence(posPair, rawCounterAll));
							supports.put(posPair, computeSupport(posPair, pairsAll));
						}
						
						// sort the pairs by confidence (a la Zimmermann et al.) (then by support for tie-breakers)
						Collections.sort(posPairs, new Comparator<LRTuple<String>>() {
							public int compare(LRTuple<String> o1, LRTuple<String> o2) {
								
								double c1 = confidences.get(o1);
								double c2 = confidences.get(o2);
								
								if (c1 > c2) {
									return +1;
								}
								else if (c1 < c2) {
									return -1;
								}
								else {
									double s1 = supports.get(o1);
									double s2 = supports.get(o2);
									
									if (s1 > s2) {
										return +1;
									}
									else if (s1 < s2) {
										return -1;
									}
									else {
										return 0;
									}
								}
							}
						});
						
						// filter out all those with less than 0.1 conf and 1 supp. (a la Zimmermann)
						posPairs = ListUtil.filterAndCopy(posPairs, new IFilter<LRTuple<String>>() {
							public boolean passesFilter(LRTuple<String> t) {
								double c = confidences.get(t);
								if (c < 0.1) {
									return false;
								}
								double s = supports.get(t);
								if (s < 1.0) {
									return false;
								}
								return true;
							}
						});
						
						// if there are more than 10, trim down to 10 (a la Zimmermann)
						if (posPairs.size() > SUGG_CUTOFF) {
							posPairs = posPairs.subList(0, SUGG_CUTOFF);
						}
						
						// otherwise, print a line that says how many more lines to expect
						pwAll.println(project + "," + trxn + "," + methodEss + "," + posPairs.size());
						
						// and then print whether each pair was realized or not
						for (LRTuple<String> posPair : posPairs) {
							// was it realized?
							if (methodsAll.contains(posPair.getLeft()) && methodsAll.contains(posPair.getRight())) {
								pwAll.println("Y");
							} else {
								pwAll.println("N");
							}
						}
					}
					
					// 2nd: update association info
					updateAssociations(pairsAll, trxn, methodsAll);
					
					// 3rd: update raw appearance info 
					updateRawCounter(rawCounterAll, trxn, methodsAll);
					
				}
				
				// do ess
				for (int trxn : trxnsAll) {
					
					if ( (trxn % 100) == 0) 
						System.out.println("Doing: " + trxn);
					
					Set<String> methodsAll = transactionsAll.get(trxn);
					Set<String> methodsEss = transactionsEss.get(trxn);
					
					if (methodsEss == null) {
						methodsEss = new HashSet<String>();
					}
					
					CollectionMap<String, LRTuple<String>> pairs = collectPosPairs(pairsEss, methodsEss);
					
					// each method is a "navigation scenario" a la ZimmermannEvaluator et al
					for (String methodEss : methodsEss)
					{
						// get all possible pairs for this method seed
						List<LRTuple<String>> posPairs = new ArrayList<LRTuple<String>>(pairs.get(methodEss));
						
						final Map<LRTuple<String>, Double> confidences = new HashMap<LRTuple<String>, Double>();
						final Map<LRTuple<String>, Double> supports = new HashMap<LRTuple<String>, Double>();
						
						// compute supports and confidences 
						for (LRTuple<String> posPair : posPairs) {
							confidences.put(posPair, computeConfidence(posPair, rawCounterEss));
							supports.put(posPair, computeSupport(posPair, pairsEss));
						}
						
						// sort the pairs by confidence (a la ZimmermannEvaluator) then by support
						Collections.sort(posPairs, new Comparator<LRTuple<String>>() {
							public int compare(LRTuple<String> o1, LRTuple<String> o2) {
								
								double c1 = confidences.get(o1);
								double c2 = confidences.get(o2);
								
								if (c1 > c2) {
									return +1;
								}
								else if (c1 < c2) {
									return -1;
								}
								else {
									double s1 = supports.get(o1);
									double s2 = supports.get(o2);
									
									if (s1 > s2) {
										return +1;
									}
									else if (s1 < s2) {
										return -1;
									}
									else {
										return 0;
									}
								}
							}
						});
						
						// filter out all those with less than 0.1 conf and 1 supp. (ZimmermannEvaluator)
						posPairs = ListUtil.filterAndCopy(posPairs, new IFilter<LRTuple<String>>() {
							public boolean passesFilter(LRTuple<String> t) {
								double c = confidences.get(t);
								if (c < 0.1) {
									return false;
								}
								double s = supports.get(t);
								if (s < 1.0) {
									return false;
								}
								return true;
							}
						});
						
						// if there are more than 5, trim down to 5 (ZimmermannEvaluator did 10, but we reduce it to 5)
						if (posPairs.size() > SUGG_CUTOFF) {
							posPairs = posPairs.subList(0, SUGG_CUTOFF);
						}
						
						// otherwise, print a line that says how many more lines to expect
						pwEss.println(project + "," + trxn + "," + methodEss + "," + posPairs.size());
						
						// and then print whether each pair was realized or not
						for (LRTuple<String> posPair : posPairs) {
							// was it realized?
							if (methodsAll.contains(posPair.getLeft()) && methodsAll.contains(posPair.getRight())) {
								pwEss.println("Y");
							} else {
								pwEss.println("N");
							}
						}
					}
					
					// 2nd: update association info
					updateAssociations(pairsEss, trxn, methodsEss);
					
					// 3rd: update raw appearance info 
					updateRawCounter(rawCounterEss, trxn, methodsEss);
				}
		
			}
			
			pwAll.close();
			pwEss.close();
		}
		catch (Exception e) {
			e.printStackTrace();
			return;
		}
		
	}

	private static void updateRawCounter(
			CollectionMap<String, Integer> rawCounterAll, int trxn,
			Set<String> methods) {
		for (String method : methods) {
			rawCounterAll.add(method, trxn);
		}
	}

	private static void updateAssociations(
			CollectionMap<LRTuple<String>, Integer> pairsAll, int trxn,
			Set<String> methods) {
		LRTupleSet<String> orderedPairs = SetUtil.getOrderedPairs(methods);
		for (LRTuple<String> methodPair : orderedPairs) {
			pairsAll.add(methodPair, trxn);
		}
	}
	
	private static double computeSupport (LRTuple<String> posPair, CollectionMap<LRTuple<String>, Integer> pairsAll) {
		return pairsAll.get(posPair).size();
	}
	
	private static double computeConfidence (LRTuple<String> posPair, CollectionMap<String, Integer> rawCounterAll)
	{
		Collection<Integer> leftTrxns = rawCounterAll.get(posPair.getLeft());
		Collection<Integer> rightTrxns = rawCounterAll.get(posPair.getRight());
		Collection<Integer> intersection = SetUtil.intersection(leftTrxns, rightTrxns);
		
		double num = intersection.size();
		double den = leftTrxns.size() + rightTrxns.size() - intersection.size();
		
		if (den == 0.0) {
			return 0.0;
		}
		
		return (num/den);
	}
	
	private static CollectionMap<String, LRTuple<String>> collectPosPairs (CollectionMap<LRTuple<String>, Integer> pairs, Set<String> methods) {
		
		CollectionMap<String, LRTuple<String>> posPairs = new CollectionMap<String, LRTuple<String>>();
		
		for (LRTuple<String> posPair : pairs.keySet()) {
			if (methods.contains(posPair.getLeft())) {
				posPairs.add(posPair.getLeft(), posPair);
			}
			if (methods.contains(posPair.getRight())) {
				posPairs.add(posPair.getRight(), posPair);
			}
		}
		
		return posPairs;
	}

	/**
	 * Takes a file containing transactions and methods touched per transaction 
	 * and returns this information as  map.
	 * @param fileName the file being analyzed 
	 */
	public static Map<Integer, Set<String>> getInfo(String fileName) {
		
		List<String> lines = Util.readLines(fileName);
		
		// 
		Set<String> allMethods = new HashSet<String>();
		Map<Integer, Set<String>> transactions = new HashMap<Integer, Set<String>>(); 
		
		// get data
		for (String line : lines) { 
			line = BindingUtil.removeGenerics(line);				// must remove generics or else split kills 			
			String[] elements = line.split(",");			 
			int id = Integer.parseInt(elements[0]);
			for (int i = 1; i < elements.length; i++) {
				allMethods.add(elements[i]);
				if (transactions.get(id) == null) {
					transactions.put(id, new HashSet<String>());
				}
				transactions.get(id).add(elements[i]);
			}
		}
		
		// trim empty information data
		Set<Integer> keys = new HashSet<Integer>(transactions.keySet());
		for (Integer key : keys) {
			if (transactions.get(key).size() <=0) {
				transactions.remove(key);
			}
		}
		return transactions;
	}
}
