package icse2011experiments.assoc;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Scanner;
import java.util.Set;

import ca.mcgill.cs.swevo.util.SetUtil;
import ca.mcgill.cs.swevo.util.filter.IFilter;

/**
 *  This file computes the statistics for Table 5 in the ICSE paper. It takes 
 *  the allAnalysis and essAnalysis files produced by AssociationChecker.java 
 *  and checks for which project,trxn,method_sig the recommendations were different. 
 *  For this intersection, it then computes the "stats" for Table 5. 
 */
public class ZimmermannEvaluator {

	public static void main(String[] args) throws IOException {
		
		final String allFile = "/home/2005/dkawry/research/experiments/icse11/weka/allAnalysis.txt";
		Set<Entry> allEntries = extractMethods(allFile);
		
		final String essFile = "/home/2005/dkawry/research/experiments/icse11/weka/essAnalysis.txt";
		Set<Entry> essEntries = extractMethods(essFile);
		
		Set<Entry> intersection = SetUtil.intersection(allEntries, essEntries);
		
		doStats(intersection);
		
		allEntries.removeAll(intersection);
		essEntries.removeAll(intersection);
		
		System.out.println("ess");
		doStats(allEntries);
		System.out.println("");
		
		System.out.println("ran");
		doStats(essEntries);
	}

    /**
     * Given a set of "Entries", computes various statistics about those entries 
     * and then prints them to system.out
     */
	private static void doStats (Set<Entry> entries) 
	{
		// the total methods that were processed
		int totalMethods = 0;

		// total methods having at least one reco
		int numFeedback = 0;

		// total number of recos that were correct ("realized" or "Y" in the file)
		int numCorrect = 0;

		// total number of recos that were incorrect ("N" in the file)
		int numIncorrect = 0;

		// number of methods that at least one "Y" reco in the top three recos
		int hadTop3 = 0;

		// number of methods that had only "N" recos 
		int totalOnlyErrs = 0;
		
		int[] posLandscape = new int[10];
		int[] negLandscape = new int[10];
		
		int numUntilFirstGood = 0;
		
		for (Entry entry : entries) {
			
			totalMethods ++;
			
			if (entry.recos.length == 0) {
				continue;
			}
			
			numFeedback ++;
			
			int tempCorrect =0;
			boolean sawOneTop3 = false;
			for (int i = 0; i < entry.recos.length; i++) {
				if (entry.recos[i]) {
					posLandscape[i] ++;
					tempCorrect ++;
					if (tempCorrect == 1) {
						numUntilFirstGood += i;
					}
				} else {
					negLandscape[i] ++;
					numIncorrect ++;
				}
				
				if (i < 3 && tempCorrect > 0) {
					sawOneTop3 = true;
				}
			}
			
			hadTop3 += sawOneTop3 ? 1 : 0;
			numCorrect += tempCorrect;
			totalOnlyErrs += tempCorrect == 0? 1 : 0;
		}
		
		System.out.println("total methods: " + totalMethods);
		System.out.println("total feedback: " + numFeedback);
		System.out.println("total correct: " + numCorrect);
		System.out.println("total incrrect: " + numIncorrect);
		System.out.println("precision: " + ((double) numCorrect)/(numCorrect + numIncorrect));
		System.out.println("total l3: " + hadTop3);
		System.out.println("ratio l3: " + ((double) hadTop3)/numFeedback);
		System.out.println("total only err: " + totalOnlyErrs);
		System.out.println("num until first good: " + numUntilFirstGood);
		System.out.println("pos landscape: " + Arrays.toString(posLandscape));
		System.out.println("neg landscape: " + Arrays.toString(negLandscape));
	}
	
	private static Set<Entry> extractMethods(final String file) throws FileNotFoundException {
		Scanner scanner = new Scanner(new File(file));
		
		Set<Entry> entries = new HashSet<Entry>();
		while (scanner.hasNextLine()) 
		{
			// 1. read the project,trxn,method,num_recos under investigation
			String line = scanner.nextLine();
			String[] parts = line.split(",");
			
			String project = parts[0].trim();
			String method = parts[2].trim();
			Integer trxn = Integer.parseInt(parts[1].trim());
			Integer numRecos = Integer.parseInt(parts[3].trim());
			boolean[] recos = new boolean[numRecos];
			
			for (int i = 0; i < numRecos; i++) {
				recos[i] = scanner.nextLine().trim().equals("Y");
			}
			
			entries.add(new Entry(project,method,trxn,recos));
		}
		
		scanner.close();
		
		return entries;
	}
	
	private static class TrueFilter implements IFilter<Boolean> {
		public boolean passesFilter(Boolean t) {
			return t;
		}
	}
	
	/**
	 * Utility class. Each Entry instance represents the recos for a given method (within a project + trxn).
     * The array "recos" is sorted in the order that the recos were presented. Index i is true if reco i was 
     * realized ("Y") in the trxn. 
     */
	private static class Entry {
		public String project;
		public String method;
		public int trxn;
		public boolean[] recos;
		
		public Entry (String project, String method, int trxn, boolean[] recos) {
			this.project = project;
			this.method = method;
			this.trxn = trxn;
			this.recos = recos;
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result
					+ ((method == null) ? 0 : method.hashCode());
			result = prime * result
					+ ((project == null) ? 0 : project.hashCode());
			result = prime * result + Arrays.hashCode(recos);
			result = prime * result + trxn;
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			Entry other = (Entry) obj;
			if (method == null) {
				if (other.method != null)
					return false;
			} else if (!method.equals(other.method))
				return false;
			if (project == null) {
				if (other.project != null)
					return false;
			} else if (!project.equals(other.project))
				return false;
			if (!Arrays.equals(recos, other.recos))
				return false;
			if (trxn != other.trxn)
				return false;
			return true;
		}
		
		
	}
}
